#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"Test typography metrics on glyph level with ReportLab."


import os
import sys
import datetime
import platform
import string
from copy import deepcopy

import reportlab
from reportlab.lib.pagesizes import landscape, A4
from reportlab.lib import colors 
from reportlab.pdfgen.canvas import Canvas
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.graphics.charts.textlabels import _text2Path
from reportlab.graphics.shapes import Drawing, Path
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.pdfmetrics import (stringWidth, getAscent, getDescent,
    registerFont, registerFontFamily)


MOVETO, LINETO, CURVETO, CLOSEPATH = range(4)
PATH_OP_ARG_COUNT = (2, 2, 6, 0)  # [moveTo, lineTo, curveTo, closePath]
    

def splitPath(p):
    "Split a path into its glyphs."
    
    ops = p.operators
    pts = p.points
    
    m = [i for (i, op) in enumerate(ops) if op == 0]
    c = [i for (i, op) in enumerate(ops) if op == 3]
    opr = zip(m, c)
    
    glyphs = [deepcopy(p) for i in range(ops.count(CLOSEPATH))]
    for i, g in enumerate(glyphs):
        g.operators = ops[slice(*opr[i])] + [3]
        s, e = [sum([PATH_OP_ARG_COUNT[ops[k]] for k in range(j)]) 
            for j in opr[i]]
        g.points = pts[s:e]
    
    return glyphs


def create_document(out_path, width, fn, verbose=False):
    "Create a document with one page per character, with paths and glyphs."

    c = Canvas(out_path, pagesize=A4) 

    c.setLineWidth(1)
    
    fs = 512.
    width = 800  
    x = 75
    y = A4[1] - fs 

    # for ch in "o":
    for ch in string.letters + string.digits: #  + u"ÄÖÜÉÈ":
        asc, dsc = getAscent(fn), getDescent(fn)
   
        # draw background box plus character
        c.setFont(fn, fs)
        d = fs*asc/1000.
        w = stringWidth(ch, fn, fs)
        c.setFillColor(colors.lightgrey)
        c.rect(x, y, w, d, stroke=0, fill=1)
        c.setFillColor(colors.black)
        c.drawString(x, y, ch)
    
        # draw base line
        c.setStrokeColor(colors.red)
        c.line(x, y, width, y)

        # draw total bounding box
        p = _text2Path(ch, fontName=fn, fontSize=fs)
        bx0, by0, bx1, by1 = p.getBounds()
        c.setStrokeColor(colors.green)
        c.rect(x + bx0, y + by0, bx1-bx0, by1-by0, stroke=1, fill=0)

        # draw character's full path contour
        p.strokeColor = colors.red
        p.fillColor = None
        p.strokeWidth = 2
        drawing = Drawing(400, 400)
        drawing.add(p)
        drawing.drawOn(c, x, y)

        # draw glyphs' bounding boxes
        glyphs = splitPath(p)
        for g in glyphs:
            bx0, by0, bx1, by1 = g.getBounds()
            c.setStrokeColor(colors.green)
            c.rect(x + bx0, y + by0, bx1-bx0, by1-by0, stroke=1, fill=0)

        # draw individual glyphs points
        radius = fs / 150.
        c.setStrokeColor(colors.green)
        # for g in [p]: 
        for g in glyphs:
            for i, op in enumerate(g.operators):
                ac = PATH_OP_ARG_COUNT[op]
                s = sum([PATH_OP_ARG_COUNT[o] for o in g.operators[:i]])
                e = sum([PATH_OP_ARG_COUNT[o] for o in g.operators[:i+1]])
                if op == MOVETO:
                    x0, y0 = g.points[s:e]
                    c.setFillColor(colors.green)
                    c.circle(x + x0, y + y0, radius*2.5, stroke=0, fill=1)
                elif op == LINETO:
                    x0, y0 = g.points[s:e]
                    c.setFillColor(colors.blue)
                    c.circle(x + x0, y + y0, radius, stroke=0, fill=1)
                elif op == CURVETO:
                    cp0x, cp0y, cp1x, cp1y, x0, y0 = g.points[s:e]
                    c.setStrokeColor(colors.orange)
                    c.setFillColor(colors.orange)
                    c.circle(x + x0, y + y0, radius, stroke=0, fill=1)
                    c.circle(x + cp0x, y + cp0y, radius, stroke=1, fill=0)
                    c.circle(x + cp1x, y + cp1y, radius, stroke=1, fill=0)
                elif op == CLOSEPATH:
                    x1, y1 = x0, y0
                    c.setFillColor(colors.red)
                    c.circle(x + x1, y + y1, radius*1.2, stroke=0, fill=1)
        c.setFillColor(colors.black)

        # write info about font, colours, versions and date/time
        fnl, fsl = "Helvetica", 12
        c.setFont(fnl, fsl)
        d = datetime.datetime.now().isoformat()
        d = d[:d.rfind(":")]
        args = (reportlab.Version, platform.python_version(), platform.platform())
        ttfLabel = ""
        try:
            if pdfmetrics._fonts[fn].__class__ == TTFont:
                ttfLabel = " (TrueType font)"
        except KeyError: 
            pass
        c.drawString(75, 100, fn + ttfLabel)
        s = "Date/Time: %s" % d 
        c.drawString(75, 80, "Dot colours (filled): green (moveto), blue " +
            "(lineto), orange (curveto), red (close)")
        s = "Reportlab v. %s, Python v. %s, Platform: %s" % args
        c.drawString(75, 60, s)
        s = "Date/Time: %s" % d 
        c.drawString(75, 40, s)

        c.showPage() 
    
    c.save()


def test():
    standardFonts = (
        'Courier', 'Courier-Bold', 'Courier-Oblique', 'Courier-BoldOblique',
        'Helvetica', 'Helvetica-Bold', 'Helvetica-Oblique', 
        'Helvetica-BoldOblique',
        'Times-Roman', 'Times-Bold', 'Times-Italic', 'Times-BoldItalic',
        'Symbol','ZapfDingbats')
    
    try:
        reportlab.rl_config.warnOnMissingFontGlyphs = 0
        registerFont(TTFont("BigCaslon", "BigCaslon.ttf"))
    except:
        print "Could not find font file. Ignoring..."

    for fn in ("Times-Roman Times-Italic Helvetica-Oblique Courier-Oblique " +
            "Symbol ZapfDingbats BigCaslon").split():        
        # fn = "Helvetica-Oblique"
        out_path = os.path.splitext(sys.argv[0])[0] + "-%s.pdf" % fn
        col = colors.black
        width = 300
        create_document(out_path, width, fn, col)


if __name__ == "__main__":
    test()
